import logging

from django.dispatch import receiver

from rest_framework_tus.models import get_upload_model
from rest_framework_tus.signals import finished, received, saved
from rest_framework_tus.storage import get_save_handler

logger = logging.getLogger(__name__)


@receiver(received, sender=get_upload_model())
def on_receiving_done(sender, instance, **kwargs):
    logger.debug(f"on_receiving_done: {instance}")
    save_handler = get_save_handler()
    save_handler(upload=instance).run()


@receiver(saved, sender=get_upload_model())
def on_saving_done(sender, instance, **kwargs):
    logger.debug(f"on_saving_done: {instance}")


@receiver(finished, sender=get_upload_model())
def on_finished(sender, instance, **kwargs):
    logger.debug(f"on_finished: {instance}")



from django.apps import AppConfig


class RestFrameworkTusConfig(AppConfig):
    name = "rest_framework_tus"

    default_auto_field = "django.db.models.BigAutoField"

    # noinspection PyUnresolvedReferences
    def ready(self):
        # Import receivers
        from . import receivers



import os

from django.conf import settings as django_settings

from dateutil import relativedelta

# Retrieve root settings dict
REST_FRAMEWORK_TUS = getattr(django_settings, "REST_FRAMEWORK_TUS", {})


# Retrieve settings
TUS_UPLOAD_MODEL = REST_FRAMEWORK_TUS.get("UPLOAD_MODEL", "rest_framework_tus.Upload")
TUS_UPLOAD_EXPIRES = REST_FRAMEWORK_TUS.get("UPLOAD_EXPIRES", relativedelta.relativedelta(days=1))
TUS_UPLOAD_DIR = REST_FRAMEWORK_TUS.get("UPLOAD_DIR", os.path.join(django_settings.BASE_DIR, "tmp", "uploads"))
TUS_UPLOAD_DESTINATION = REST_FRAMEWORK_TUS.get("TUS_UPLOAD_DESTINATION", "uploaded")
TUS_RESPONSE_BODY_ENABLED = REST_FRAMEWORK_TUS.get("RESPONSE_BODY_ENABLED", False)
TUS_SAVE_HANDLER_CLASS = REST_FRAMEWORK_TUS.get("SAVE_HANDLER_CLASS", "rest_framework_tus.storage.DefaultSaveHandler")
TUS_MAX_FILE_SIZE = REST_FRAMEWORK_TUS.get("MAX_FILE_SIZE", 4 * 1024 * 1024 * 1024)  # 4 GB
TUS_FILENAME_METADATA_FIELD = REST_FRAMEWORK_TUS.get("FILENAME_METADATA_FIELD", "filename")



from django.utils.translation import gettext_lazy as _

from rest_framework import status
from rest_framework.exceptions import APIException


class Conflict(APIException):
    status_code = status.HTTP_409_CONFLICT
    default_detail = _("Conflict.")
    default_code = "conflict"



import rest_framework
from rest_framework.routers import DynamicRoute, Route, SimpleRouter


def get_list_route():
    list_route_data = dict(
        url=r"^{prefix}{trailing_slash}$",
        mapping={
            "get": "list",
            "post": "create",
        },
        name="{basename}-list",
        detail=False,
        initkwargs={"suffix": "List"},
    )

    return Route(**list_route_data)


def get_detail_route():
    detail_route_data = dict(
        url=r"^{prefix}/{lookup}{trailing_slash}$",
        mapping={
            "get": "retrieve",
            "put": "update",
            "patch": "partial_update",
            "delete": "destroy",
            "head": "info",
        },
        name="{basename}-detail",
        detail=False,
        initkwargs={"suffix": "Instance"},
    )

    return Route(**detail_route_data)


class TusAPIRouter(SimpleRouter):
    routes = [
        # List route.
        get_list_route(),
        # Dynamically generated list routes.
        # Generated using @list_route decorator
        # on methods of the viewset.
        DynamicRoute(
            url=r"^{prefix}/{methodname}{trailing_slash}$",
            name="{basename}-{methodnamehyphen}",
            detail=False,
            initkwargs={},
        ),
        # Detail route.
        get_detail_route(),
        # Dynamically generated detail routes.
        # Generated using @detail_route decorator on methods of the viewset.
        DynamicRoute(
            url=r"^{prefix}/{lookup}/{methodname}{trailing_slash}$",
            name="{basename}-{methodnamehyphen}",
            detail=True,
            initkwargs={},
        ),
    ]



# Generated by Django 1.10.6 on 2017-03-09 13:24

import uuid

from django.db import migrations, models

import django_fsm
import jsonfield.fields


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Upload",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("guid", models.UUIDField(default=uuid.uuid4, unique=True, verbose_name="GUID")),
                ("state", django_fsm.FSMField(default="initial", max_length=50)),
                ("upload_offset", models.PositiveIntegerField(default=0)),
                ("upload_length", models.IntegerField(default=-1)),
                ("upload_metadata", jsonfield.fields.JSONField()),
                ("filename", models.CharField(blank=True, max_length=255)),
                ("temporary_file_path", models.CharField(max_length=4096, null=True)),
                ("expires", models.DateTimeField(blank=True, null=True)),
                ("uploaded_file", models.FileField(upload_to="uploaded")),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django.http.response import HttpResponse

from rest_framework import status

from . import constants, tus_api_version
from .compat import decode_base64


class TusMiddleware:
    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = None
        if hasattr(self, "process_request"):
            response = self.process_request(request)
        if not response:
            response = self.get_response(request)
        if hasattr(self, "process_response"):
            response = self.process_response(request, response)
        return response

    def process_request(self, request):
        # Parse tus client version
        self.parse_tus_version(request)

        # Parse upload length
        self.parse_upload_length(request)

        # Parse upload upload_offset
        self.parse_upload_offset(request)

        # Parse defer upload length
        self.parse_upload_defer_length(request)

        # Parse upload metadata
        self.parse_upload_metadata(request)

        # Parse upload checksum
        self.parse_upload_checksum(request)

    def process_response(self, request, response):
        if "Tus-Resumable" not in response:
            response["Tus-Resumable"] = tus_api_version

        return response

    @classmethod
    def parse_tus_version(cls, request):
        tus_version = cls.get_header(request, "Tus-Resumable", None)

        if tus_version is None:
            return

        # Set upload length
        setattr(request, constants.TUS_RESUMABLE_FIELD_NAME, tus_version)

    @classmethod
    def parse_upload_defer_length(cls, request):
        upload_defer_length = cls.get_header(request, "Upload-Defer-Length", None)

        if not upload_defer_length:
            return

        upload_defer_length = int(upload_defer_length)

        if upload_defer_length != 1:
            return HttpResponse(
                f'Invalid value for "Upload-Defer-Length" header: {upload_defer_length}.',
                status=status.HTTP_400_BAD_REQUEST,
            )

        # Set upload defer length
        setattr(request, constants.UPLOAD_DEFER_LENGTH_FIELD_NAME, upload_defer_length)

    @classmethod
    def parse_upload_offset(cls, request):
        upload_offset = cls.get_header(request, "Upload-Offset", None)

        if upload_offset is None:
            return

        # Set upload length
        setattr(request, constants.UPLOAD_OFFSET_NAME, int(upload_offset))

    @classmethod
    def parse_upload_length(cls, request):
        upload_length = cls.get_header(request, "Upload-Length", None)

        if upload_length is None:
            return

        # Set upload length
        setattr(request, constants.UPLOAD_LENGTH_FIELD_NAME, int(upload_length))

    @classmethod
    def parse_upload_checksum(cls, request):
        upload_checksum_header = cls.get_header(request, "Upload-Checksum", None)

        if upload_checksum_header is None:
            return

        upload_checksum = list(upload_checksum_header.split(" "))
        if len(upload_checksum) != 2:
            return HttpResponse(
                f'Invalid value for "Upload-Checksum" header: {upload_checksum_header}.',
                status=status.HTTP_400_BAD_REQUEST,
            )

        # Set upload checksum
        setattr(request, constants.UPLOAD_CHECKSUM_FIELD_NAME, upload_checksum)

    @classmethod
    def parse_upload_metadata(cls, request):
        upload_meta_header = cls.get_header(request, "Upload-Metadata", None)

        if upload_meta_header is None:
            return

        upload_metadata = {}

        for key_value_pair in upload_meta_header.split(","):
            # Trim whitespace
            key_value_pair = key_value_pair.strip()

            # Split key and value
            key, value = key_value_pair.split(" ")

            # Store data
            upload_metadata[key] = decode_base64(value.encode("ascii")).decode("utf-8")

        # Set upload_metadata
        setattr(request, constants.UPLOAD_METADATA_FIELD_NAME, upload_metadata)

    @classmethod
    def get_header(cls, request, key, default_value=None):
        # First, we try to retrieve the key in the "headers" dictionary
        result = request.META.get("headers", {}).get(key, None)

        # If we didn't find the key, or the value was "None", try to use the "HTTP_{uppercased-key}" key
        if result is None:
            custom_value = "HTTP_{}".format(key.replace("-", "_").upper())
            result = request.META.get(custom_value, default_value)

        # If we didn't find the key, or the value was "None", try to use the "HTTP_X_{uppercased-key}" key
        if result is None:
            # https://tools.ietf.org/html/rfc6648
            custom_value = "HTTP_X_{}".format(key.replace("-", "_").upper())
            result = request.META.get(custom_value, default_value)

        # If we still didn't find the key, or the value was "None", return the default value
        if result is None:
            result = default_value

        # Return the result
        return result



# Generated by Django 1.10.6 on 2017-04-06 11:20

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("rest_framework_tus", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="upload",
            name="uploaded_file",
            field=models.FileField(blank=True, max_length=255, null=True, upload_to="uploaded"),
        ),
    ]



# Generated by Django 1.11.2 on 2017-06-19 08:58

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("rest_framework_tus", "0002_auto_20170406_0620"),
    ]

    operations = [
        migrations.AlterField(
            model_name="upload",
            name="upload_length",
            field=models.BigIntegerField(default=-1),
        ),
        migrations.AlterField(
            model_name="upload",
            name="upload_offset",
            field=models.BigIntegerField(default=0),
        ),
    ]



from django.urls import include, path

from rest_framework_tus.views import UploadViewSet

from .routers import TusAPIRouter

app_name = "rest_framework_tus"

router = TusAPIRouter()
router.register(r"files", UploadViewSet, basename="upload")

urlpatterns = [
    path(r"", include((router.urls, "rest_framework_tus"), namespace="api")),
]



__version__ = "2.0.1"

tus_api_version = "1.0.0"
tus_api_version_supported = ["1.0.0"]
tus_api_extensions = ["creation", "creation-defer-length", "termination", "checksum", "expiration"]
tus_api_checksum_algorithms = ["md5", "sha1", "sha224", "sha256", "sha384", "sha512"]



# Generated by Django 4.1 on 2022-12-17 07:53

import collections

from django.db import migrations, models

import jsonfield.fields


class Migration(migrations.Migration):
    dependencies = [
        ("rest_framework_tus", "0003_auto_20170619_0358"),
    ]

    operations = [
        migrations.AlterField(
            model_name="upload",
            name="id",
            field=models.BigAutoField(
                auto_created=True,
                primary_key=True,
                serialize=False,
                verbose_name="ID",
            ),
        ),
        migrations.AlterField(
            model_name="upload",
            name="upload_metadata",
            field=jsonfield.fields.JSONField(
                load_kwargs={"object_pairs_hook": collections.OrderedDict},
            ),
        ),
    ]



TUS_RESUMABLE_FIELD_NAME = "tus_resumable"

UPLOAD_DEFER_LENGTH_FIELD_NAME = "tus_upload_defer_length"
UPLOAD_LENGTH_FIELD_NAME = "tus_upload_length"
UPLOAD_OFFSET_NAME = "tus_upload_offset"
UPLOAD_METADATA_FIELD_NAME = "tus_upload_metadata"
UPLOAD_CHECKSUM_FIELD_NAME = "tus_upload_checksum"



from rest_framework.parsers import BaseParser, DataAndFiles


class TusUploadStreamParser(BaseParser):
    media_type = "application/offset+octet-stream"

    def parse(self, stream, media_type=None, parser_context=None):
        return DataAndFiles({"chunk": stream.body}, {})



# Generated by Django 4.2.2 on 2023-07-04 06:56

from django.db import migrations, models

import rest_framework_tus.models


class Migration(migrations.Migration):
    dependencies = [
        ("rest_framework_tus", "0004_alter_upload_id_alter_upload_upload_metadata"),
    ]

    operations = [
        migrations.AlterField(
            model_name="upload",
            name="uploaded_file",
            field=models.FileField(
                blank=True,
                max_length=255,
                null=True,
                upload_to=rest_framework_tus.models.custom_upload_path,
            ),
        ),
    ]



#!/usr/bin/env python
import os
import re
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


def get_version(*file_paths):
    """Retrieves the version from rest_framework_tus/__init__.py"""
    filename = os.path.join(os.path.dirname(__file__), *file_paths)
    version_file = open(filename).read()
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file,
        re.M,
    )
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


version = get_version("rest_framework_tus", "__init__.py")


if sys.argv[-1] == "publish":
    try:
        import wheel

        print("Wheel version: ", wheel.__version__)
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    sys.exit()

if sys.argv[-1] == "tag":
    print("Tagging the version on git:")
    os.system(f"git tag -a {version} -m 'version {version}'")
    os.system("git push --tags")
    sys.exit()

readme = open("README.rst").read()
history = open("HISTORY.rst").read().replace(".. :changelog:", "")

setup(
    name="drf-tus",
    version=version,
    description="A Tus (tus.io) library for Django Rest Framework",
    long_description=f"{readme}\n\n{history}",
    author="Dirk Moors",
    author_email="dirkmoors@gmail.com",
    url="https://github.com/dirkmoors/drf-tus",
    packages=[
        "rest_framework_tus",
    ],
    include_package_data=True,
    install_requires=[
        "python-dateutil>=2.8.2",
        "Django>=3.2,<=4.2",
        "djangorestframework>=3.14.0",
        "jsonfield>=2.0.0",
        "django-fsm==2.8.1",
        "six>=1.11.0",
    ],
    license="MIT",
    zip_safe=False,
    keywords=[
        "drf-tus",
        "tus",
        "django",
        "rest",
        "framework",
        "django-rest-framework",
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        "Framework :: Django :: 4.1",
        "Framework :: Django :: 4.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
